import matplotlib.pyplot as plt
from matplotlib import cm
from mpl_toolkits.mplot3d import axes3d
import seaborn as sns
import numpy as np
import pandas as pd
import copy
import itertools
            
    
def plot_main(ga2m, X, column_names=None):
    num_use = len(ga2m.use_main_features)
    if num_use < 3:
        num_col = num_use
    else:
        num_col = 3
    num_row = int((num_use + 2) / 3)

    width = num_col * 4
    height = num_row * 3
    fig, axs = plt.subplots(nrows=num_row, ncols=num_col, figsize=(width, height))
    axs = axs.reshape(-1)

    for p, i in enumerate(ga2m.use_main_features):
        max_x = np.max(X[:, i])
        min_x = np.min(X[:, i])
        x = np.linspace(start=min_x, stop=max_x)
        a = ga2m.main_model_dict[i].predict(x.reshape(-1, 1), num_iteration=ga2m.main_model_dict[i].best_iteration)

        if type(column_names) != type(None):
            axs[p].set_title(column_names[i])
        else:
            axs[p].set_title(i)
        axs[p].set_xlabel(r'$x_{}$'.format(i), fontsize=12)
        axs[p].set_ylabel(r'$f_{}(x_{})$'.format(i, i), fontsize=12)
        sns.lineplot(x=x, y=a, ax=axs[p])

    plt.tight_layout()
    plt.show()

    
    
def plot_interaction(ga2m,X,mode = '3d'):
    if mode == '3d':
        plot_interaction_3d(ga2m,X)
        
    else:
        num_use = len(ga2m.use_interaction_features)
        if num_use < 3:
            num_col = num_use
        else:
            num_col = 3
        num_row = int((num_use+2)/3)

        width = num_col * 4
        hight = num_row * 3
        fig, axs = plt.subplots(nrows=num_row,ncols=num_col,figsize=(width,hight))
        axs = axs.reshape(-1)

        for p,(i,j) in enumerate(ga2m.use_interaction_features):
            #カテゴリカルと離散数値の場合はあとで対処
            max_x0 = np.max(X[:,i])
            min_x0 = np.min(X[:,i])
            max_x1 = np.max(X[:,j])
            min_x1 = np.min(X[:,j])
            x0 = np.linspace(start=min_x0,stop=max_x0)
            x1 = np.linspace(start=min_x1,stop=max_x1)

            a,b =np.meshgrid(x0,x1)
            x =np.hstack((a.reshape(-1,1),b.reshape(-1,1)))
            preds = ga2m.interaction_model_dict[(i,j)].predict(x.reshape(-1,2),num_iteration=ga2m.interaction_model_dict[(i,j)].best_iteration)


            cp = axs[p].contourf(a, b, preds.reshape(a.shape))
            plt.colorbar(cp,ax=axs[p])

            axs[p].set_title(r'$f_{0}._{1}(x_{0},x_{1})$'.format(i,j))
            axs[p].set_xlabel(r'$x_{}$'.format(i))
            axs[p].set_ylabel(r'$x_{}$'.format(j))
        plt.tight_layout()
        plt.show()

    
    
def show_importance(ga2m,after_prune=True,higher_mode=False):
    # 重要度(←これはおまけ)
    if after_prune:
        tmp_dict = copy.deepcopy(ga2m.after_feature_importance_)
    else:
        tmp_dict = copy.deepcopy(ga2m.before_feature_importance_)

    if higher_mode == False:
        del tmp_dict['higher']

    impact_df = pd.DataFrame(tmp_dict.values(),
                             index=[str(a) for a in tmp_dict.keys()],
                             columns=['IMPORTANCE'])
    sns.barplot(data=impact_df.T, orient='h')
    plt.title('IMPORTANCE')
    plt.show()


def plot_interaction_3d(ga2m,X):
    num_use = len(ga2m.use_interaction_features)
    if num_use < 3:
        num_col = num_use
    else:
        num_col = 3
    num_row = int((num_use+2)/3)
    # print(num_col, num_row)
    width = num_col * 4
    hight = num_row * 3
    fig = plt.figure(figsize=(width,hight))

    for p,(i,j) in enumerate(ga2m.use_interaction_features):
        #カテゴリカルと離散数値の場合はあとで対処
        axs = fig.add_subplot(num_row, num_col, p+1, projection='3d')
        max_x0 = np.max(X[:,i])
        min_x0 = np.min(X[:,i])
        max_x1 = np.max(X[:,j])
        min_x1 = np.min(X[:,j])
        x0 = np.linspace(start=min_x0,stop=max_x0)
        x1 = np.linspace(start=min_x1,stop=max_x1)

        a,b =np.meshgrid(x0,x1)
        x =np.hstack((a.reshape(-1,1),b.reshape(-1,1)))
        preds = ga2m.interaction_model_dict[(i,j)].predict(x.reshape(-1,2),num_iteration=ga2m.interaction_model_dict[(i,j)].best_iteration)


        cp = axs.plot_surface(a, b, preds.reshape(a.shape), rstride=1, cstride=1, cmap=cm.coolwarm)
        #plt.colorbar(cp,ax=axs)

        axs.set_title(r'$f_{0}._{1}(x_{0},x_{1})$'.format(i,j))
        axs.set_xlabel(r'$x_{}$'.format(i))
        axs.set_ylabel(r'$x_{}$'.format(j))
    plt.tight_layout()
    plt.show()

    
"""
# カテゴリ変数のプロット
def plot_cat(ga2m, X, column_names = None, categorical_features=[]):
    categorical_features = list(set(ga2m.use_main_features)&set(categorical_features))
    num_cat = len(categorical_features)
    if num_cat < 3:
        num_col = num_cat
    else:
        num_col = 3
    num_row = int((num_cat+2)/3)
    
    width = num_col * 4
    height = num_row * 3
    fig, axs = plt.subplots(nrows = num_row, ncols = num_col, figsize=(width, height))
    axs = axs.reshape(-1)
    
    for p, i in enumerate(categorical_features):
        n_unique = np.unique(X[:,i])
        a = ga2m.main_model_dict[i].predict(n_unique.reshape(-1,1))
        if type(column_names) != type(None):
            axs[p].set_title(column_names[i])
        else:
            axs[p].set_title(i)
        axs[p].set_xlabel(r"$x_{}$".format(i), fontsize=12)
        axs[p].set_ylabel(r"$f(x_{})$".format(i), fontsize=12)
        axs[p].bar(n_unique, a)
        
    plt.tight_layout()
    plt.show()

    
# カテゴリ変数×カテゴリ変数のプロット
def plot_cat_interaction(ga2m, X, column_names = None, categorical_features=[]):
    categorical_features = list(set(ga2m.use_main_features)&set(categorical_features))
    categorical_features = sorted(categorical_features)
    combinations = list(itertools.combinations(categorical_features,2))
    categorical_features = list(set(ga2m.use_interaction_features)&set(combinations))    
    
    num_cat = len(categorical_features)
    if num_cat < 3:
        num_col = num_cat
    else:
        num_col = 3
    num_row = int((num_cat+2)/3)
    
    width = num_col * 4
    height = num_row * 3
    fig, axs = plt.subplots(nrows=num_row, ncols=num_col, figsize=(width, height))
    if num_cat > 1:
        axs = axs.reshape(-1)

    for p, (i,j) in enumerate(categorical_features):
        x0 = np.unique(X[:,i])
        x1 = np.unique(X[:,j])
        a,b =np.meshgrid(x0,x1)
        x =np.hstack((b.reshape(-1,1),a.reshape(-1,1)))
        preds = ga2m.interaction_model_dict[(i,j)].predict(x.reshape(-1,2),num_iteration=ga2m.interaction_model_dict[(i,j)].best_iteration)
        preds = preds.reshape(-1,1)
        data = np.hstack((x,preds))
        data = pd.DataFrame(data, columns = ["a", "b", "c"])
        data = data.pivot("a", "b", "c")
        if num_cat > 1:
            sns.heatmap(data, annot=False, linewidths = 0.7, ax = axs[p])
            axs[p].set_title(r'$f_{0}._{1}(x_{0},x_{1})$'.format(i,j))
            axs[p].set_xlabel(r'$x_{}$'.format(i))
            axs[p].set_ylabel(r'$x_{}$'.format(j))
        else:
            sns.heatmap(data, annot=False, linewidths = 0.7, ax = axs)
            axs.set_title(r'$f_{0}._{1}(x_{0},x_{1})$'.format(i,j))
            axs.set_xlabel(r'$x_{}$'.format(i))
            axs.set_ylabel(r'$x_{}$'.format(j))
    plt.tight_layout()
    plt.show()


# 連続変数×カテゴリ変数のプロット
def plot_catcon_interaction(ga2m, X, column_names = None, categorical_features=[]):
    ## モデル内で使用されているカテゴリ変数、連続変数を抽出
    categorical_features = sorted(list(set(ga2m.use_main_features)&set(categorical_features)))
    continuous_features = sorted(list(set(ga2m.use_main_features)^set(categorical_features)))
    
    ## カテゴリ変数、連続変数の組み合わせ
    cat_combinations = list(itertools.combinations(categorical_features, 2))
    con_combinations = list(itertools.combinations(continuous_features, 2))
    ## モデル内で使用されているカテゴリ変数、連続変数の組み合わせを抽出
    cat_combinations = list(set(ga2m.use_interaction_features)&set(cat_combinations))
    con_combinations = list(set(ga2m.use_interaction_features)&set(con_combinations))
    cat_con_combinations = list(set(cat_combinations)|set(con_combinations))
    ## モデル内で使用されているカテゴリ変数×連続変数の組み合わせを抽出
    catcon_combinations = list(set(ga2m.use_interaction_features)-set(cat_con_combinations))
    
    num_cat = len(catcon_combinations)
    if num_cat < 2:
        num_col = num_cat
    else:
        num_col = 2
    num_row = int((num_cat+1)/2)
    
    width = num_col * 6
    height = num_row * 4
    fig, axs = plt.subplots(nrows=num_row, ncols=num_col, figsize=(width, height))
    axs = axs.reshape(-1)
    
    for p, (i,j) in enumerate(catcon_combinations):
        cat_unique = sorted(list((np.unique(X[:,j]))))
        for u in cat_unique:
            x = np.empty((1000, 2))
            x[:,0] = np.linspace(X[:,i].min(), X[:,i].max(), 1000)
            x[:,1] = u
            preds =ga2m.interaction_model_dict[(i,j)].predict(x.reshape(-1,2),num_iteration=ga2m.interaction_model_dict[(i,j)].best_iteration)
            axs[p].plot(x[:,0], preds, label=r"$x_{0}$={1}".format(j,u), linewidth=0.7)
        
        axs[p].legend(loc="best",fontsize=8)
        axs[p].set_title(r'$f_{0}._{1}(x_{0},x_{1})$'.format(i,j))
        axs[p].set_xlabel(r'$x_{}$'.format(i))
        axs[p].set_ylabel(r'$f_{0}._{1}(x_{0},x_{1})$'.format(i,j))
    plt.tight_layout()
    plt.show()

"""


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cga2m_plus', 'cga2m_plus..ipynb_checkpoints']

package_data = \
{'': ['*']}

install_requires = \
['lightgbm>=3.2.1,<4.0.0',
 'matplotlib>=3.4.2,<4.0.0',
 'numpy',
 'seaborn>=0.11.1,<0.12.0']

setup_kwargs = {
    'name': 'cga2m-plus',
    'version': '0.1.5',
    'description': 'CGA2M_plus is a Python package of Constraint GA2M plus(CGA2M+). CGA2M+ is a model that improves GA2M in two respects: interpretability and accuracy.',
    'long_description': '![](https://raw.githubusercontent.com/MK-tech20/CGA2M_plus/main/images/cga2m_plus%2B.png) \n# CGA2M+ (Constraint GA2M plus)\nWe propose Constraint GA2M plus (CGA2M+), which we proposed. CGA2M+ is a modified version of GA2M to improve its interpretability and accuracy.\nFor more information, please read our paper.(coming soon!!) \nMainly, CGA2M+ differs from GA2M in two respects.\n1. introducing monotonic constraints\n2. introducing higher-order interactions keeping the interpretability of the model\n# Description of CGA2M+\nMainly, CGA2M+ differs from GA2M in two respects. We are using LightGBM as a shape function.\n\n- introducing monotonic constraints\nBy adding monotonicity, we can improve the interpretability of our model. For example, we can make sure that "in the real estate market, as the number of rooms increases, the price decreases" does not happen. Human knowledge is needed to determine which features to enforce monotonicity on. The monotonicity constraint algorithm is implemented in LightGBM. This is a way to constrain the branches of a tree. For more details, please refer to the LightGBM implementation.\n\n![](https://raw.githubusercontent.com/MK-tech20/CGA2M_plus/main/images/constraint.png)   \n\n- introducing higher-order interactions keeping the interpretability of the model\nGGA2M is unable to take into account higher-order interactions. Therefore, we introduce higher-order terms that are not interpretable. However, we devise a learning method so that the higher-order terms do not compromise the overall interpretability. Specifically, we train the higher-order terms as models that predict the residuals of the univariate terms and pairwise interaction terms. This allows most of the predictions to be explained by the interpretable first and second order terms. These residuals are then predicted by a higher-order term.\n\n# Algorithm  \n![](https://raw.githubusercontent.com/MK-tech20/CGA2M_plus/main/images/algorithm.png)\nFor more information, please read our paper. (coming soon!!) \n# Installation\nYou can get CGA2M+ from PyPI. Our project in PyPI is [here](https://pypi.org/project/cga2m-plus/).\n```bash\npip install cga2m-plus\n```\n\n# Usage\nFor more detail, please read `examples/How_to_use_CGA2M+.ipynb`.\nIf it doesn\'t render at all in github, please click [here](https://kokes.github.io/nbviewer.js/viewer.html#aHR0cHM6Ly9naXRodWIuY29tL01LLXRlY2gyMC9DR0EyTV9wbHVzL2Jsb2IvbWFpbi9leGFtcGxlcy9Ib3dfdG9fdXNlX0NHQTJNJTJCLmlweW5i).\n## Training\n\n```python\ncga2m = Constraint_GA2M(X_train,\n                        y_train,\n                        X_eval,\n                        y_eval,\n                        lgbm_params,\n                        monotone_constraints = [0] * 6,\n                        all_interaction_features = list(itertools.combinations(range(X_test.shape[1]), 2)))\n\ncga2m.train(max_outer_iteration=20,backfitting_iteration=20,threshold=0.05)\ncga2m.prune_and_retrain(threshold=0.05,backfitting_iteration=30)\ncga2m.higher_order_train()\n```\n## Predict\n```python\ncga2m.predict(X_test,higher_mode=True)\n```\n\n## Visualize the effect of features on the target variables.\n```python\nplot_main(cga2m_no1,X_train)\n```\n![](https://raw.githubusercontent.com/MK-tech20/CGA2M_plus/main/images/plot_main.png) \n\n## Visualize (3d) the effect of pairs of features on the target variables\n```python\nplot_interaction(cga2m_no1,X_train,mode = "3d")\n```\n![](https://raw.githubusercontent.com/MK-tech20/CGA2M_plus/main/images/plot_pairs.png) \n## Feature importance\n```python\nshow_importance(cga2m_no1,after_prune=True,higher_mode=True)\n```\n![](https://raw.githubusercontent.com/MK-tech20/CGA2M_plus/main/images/feature_importance.png) \n# License\nMIT License\n# Citation\nYou may use our package(CGA2M+) under MIT License. \nIf you use this program in your research then please cite:\n\n**CGA2M+ Package**  \n```bash\nM. Kuramata, A. Watanabe, K. Majima, H. Kiyohara, M. Iwata, K. Kondo, K. Nakata, \nCGA2M+ Package. (2021) [Online].\nAvailable: https://github.com/MK-tech20/CGA2M_plus\n```\n\n**CGA2M+ Paper**  \nThe paper has not been published yet. \n\n# Reference\n[1] Friedman, J. H. 2001, Greedy function approximation: a gradient boosting machine, Annals of statistics, 1189-1232, doi: 10.1214/aos/1013203451. Available online: May 02, 2021\n\n[2] Ke, G., Meng, Q., Finley, T., Wang, T., Chen, W., Ma, W., ... Liu, T. Y. 2017. Lightgbm: A highly efficient gradient boosting decision tree, Advances in neural information processing systems(NIPS’17), Long Beach California , 4-9 December, pp. 3146-3154.\n\n[3] Nelder, J. A., Wedderburn, R. W. 1972. Generalized linear models, Journal of the Royal Statistical Society: Series A (General), 135(3), 370-384, doi: 10.2307/2344614, Available online: May 02, 2021\n\n[4] Hastie, T. J., Tibshirani, R. J. 1990. Generalized additive models (Vol. 43), CRC press, doi: 10.1214/ss/1177013604. Available online: May 02, 2021\n\n[5] Lou, Y., Caruana, R., Gehrke, J., Hooker, G. 2013, August. Accurate intelligible models with pairwise interactions, Proceedings of the 19th ACM SIGKDD international conference on Knowledge discovery and data mining(KDD’13), Chicago Illinois, United States of America, 11-14 August, pp. 623-631.\n\n[6] “GitHub - microsoft/LightGBM” [Online]. Available: https://github.com/microsoft/LightGBM (Accessed: May 02, 2021)\n\n[7] “scikit-learn: machine learning in Python — scikit-learn 0.24.2 documentation” [Online]. Available: https://scikit-learn.org/stable/ (Accessed May 02, 2021)\n',
    'author': 'M. Kuramata',
    'author_email': 'baisgtud43j@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MK-tech20',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
